import sys
import unittest
import numpy as np

sys.path.append('..')

from pysprint.core.dataedits import *

class TestEdit(unittest.TestCase):

	def setUp(self):
		pass

	def tearDown(self):
		pass

	def test_savgol(self):
		x, y, v, w = np.loadtxt('test_arms.txt', delimiter = ',', unpack = True)
		a, b = savgol(x, y, v, w, window = 10, order = 3)
		c, d = savgol(x, y, v, w, window = 11, order = 3)
		assert len(a) == len(b)
		assert len(d) == len(d)
		with self.assertRaises(ValueError):
			savgol(x, y, v, w, window = 1, order = 3)


	
	def test_peak(self):
		x, y = np.loadtxt('test_peak.txt', delimiter = ',', unpack = True)
		a,b,c,d = find_peak(x, y, [], [], threshold = 0.01, proMin = 0.5, proMax = 0.5)
		assert len(a) == len(b)
		assert len(c) == len(d)
		for val in b:
			assert abs(val) > 0.01
		for val in d:
			assert abs(val) > 0.01

	def test_interpolate(self):
		x, y, v, w = np.loadtxt('test_arms.txt', delimiter = ',', unpack = True)
		a, b = interpolate_data(x, y, v, w)
		assert len(a) == len(b)
		np.random.seed(1000)
		idx1 = np.random.randint(0, len(a))
		idx2 = np.random.randint(0, len(a))
		i = abs(a[idx1] - a[idx1-1])
		j = abs(a[idx2] - a[idx2-1])
		np.testing.assert_almost_equal(i, j)
		with self.assertRaises(TypeError):
			interpolate_data({'a': '1', 'b':'2'}, x,y,v, window = 1, order = 3)

	def test_cut(self):
		x, y, v, w = np.loadtxt('test_arms.txt', delimiter = ',', unpack = True)
		a, b = cut_data(x, y, v, w, startValue = 2.2, endValue = 2.8)
		assert len(a) == len(b)
		np.testing.assert_almost_equal(min(a), 2.2, decimal = 2)
		np.testing.assert_almost_equal(max(a), 2.8, decimal = 2)
		with self.assertRaises(ValueError):
			cut_data(x, [], v, w)


	def test_convolution(self):
		x, y, v, w = np.loadtxt('test_arms.txt', delimiter = ',', unpack = True)
		a, b = convolution(x, y, v, w, standev = 200)
		assert len(a) == len(b)
		with self.assertRaises(ValueError):
			convolution(x, [], v, w, standev = -541)

if __name__ == '__main__':
	unittest.main()
		



"""
Sample generator
"""
import numpy as np 	
# from datetime import datetime

C_LIGHT = 299.793 #nm/fs


def _ensure_input(start, stop, center, resolution, pulseWidth):
	if start >= stop:
		raise ValueError('start value must be less than stop')
	if center < start or center > stop:
		raise ValueError('center must be between start and stop')	
	if resolution > (stop-start):
		raise ValueError('resolution is too big')
	if pulseWidth <= 0:
		raise ValueError('Pulse width is strictly positive.')
	else:
		pass


def _disp(x, delay, GD=0, GDD=0, TOD=0, FOD=0, QOD=0):
	return x*GD+x*delay+(GDD/2)*x**2+delay+(TOD/6)*x**3+(FOD/24)*x**4+(QOD/120)*x**5


def generatorFreq(start, stop, center, delay, GD=0, GDD=0, TOD=0, FOD=0, QOD=0, resolution=0.1,
				  delimiter=',',pulseWidth=10, includeArms=False, chirp=0):
	_ensure_input(start, stop, center, resolution, pulseWidth)
	omega0 = center
	window = (np.sqrt(1+chirp**2)*8*np.log(2))/(pulseWidth**2)
	lamend = (2*np.pi*C_LIGHT)/start
	lamstart = (2*np.pi*C_LIGHT)/stop
	# stepAmount = (lamend-lamstart+resolution)/resolution
	# lam = np.linspace(lamstart, lamend+resolution,stepAmount)
	lam = np.arange(lamstart, lamend+resolution, resolution) 
	omega = (2*np.pi*C_LIGHT)/lam 
	relom = omega-omega0
	i1 = np.exp(-(relom)**2/(window))
	i2 = np.exp(-(relom)**2/(window))
	i = i1 + i2 + 2*np.sqrt(i1*i2)*np.cos(_disp(relom, delay, GD=GD, GDD=GDD, TOD=TOD, FOD=FOD, QOD=QOD)+(2*delay*omega/C_LIGHT))
	if includeArms:
		return omega, i, i1, i2
		# np.savetxt('examples/simulated_'+str(datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))+'_frequency.txt', np.transpose([omega ,i, i1, i2]), 
			# header = 'freq, int, ref, sam', delimiter = delimiter, comments ='')
	else:
		# np.savetxt('examples/simulated_'+str(datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))+'_frequency.txt', np.transpose([omega ,i,]), 
			# header = 'freq, int', delimiter = delimiter, comments ='')
		return omega, i, np.array([]), np.array([])

#TODO: intenzitásarány

def generatorWave(start, stop, center, delay, GD=0, GDD=0, TOD=0, FOD=0, QOD=0, resolution=0.1, 
				  delimiter=',',pulseWidth=10, includeArms=False, chirp=0):
	_ensure_input(start, stop, center, resolution, pulseWidth)
	omega0 = (2*np.pi*C_LIGHT)/center 
	window = (np.sqrt(1+chirp**2)*8*np.log(2))/(pulseWidth**2)
	# stepAmount = (stop-start+resolution)/resolution
	# lam = np.linspace(start, stop+resolution, stepAmount)
	lam = np.arange(start, stop+resolution, resolution) 
	omega = (2*np.pi*C_LIGHT)/lam
	relom = omega-omega0 
	i1 = np.exp(-(relom)**2/(window))
	i2 = np.exp(-(relom)**2/(window))
	i = i1 + i2 + 2*np.cos(_disp(relom, delay, GD=GD, GDD= GDD, TOD=TOD, FOD=FOD, QOD=QOD)+(2*omega*delay/C_LIGHT))*np.sqrt(i1*i2)
	if includeArms:
		return lam, i, i1, i2
		# np.savetxt('examples/simulated_'+str(datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))+'_wavelength.txt', np.transpose([lam ,i, i1, i2]), 
			# header = 'wavelength, int, ref, sam', delimiter = delimiter, comments ='')
	else:
		return lam, i, np.array([]), np.array([])
		# np.savetxt('examples/simulated_'+str(datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))+'_wavelength.txt', np.transpose([lam ,i]), 
			# header = 'wavelength, int', delimiter = delimiter, comments ='')




def main():
    print('Building up UI..')
    import sys
    from pysprint.logic import MainProgram
    try:
        from PyQt5 import QtWidgets
    except ImportError:
        print('PyQt5 is essential for the UI. You can use the API instead.')
    app = QtWidgets.QApplication(sys.argv)
    my_interface = MainProgram()
    my_interface.showMaximized()
    if my_interface.settings.value('show') == 'True':
    	my_interface.msgbox.exec_()
    	if my_interface.cb.isChecked():
           my_interface.settings.setValue('show', False)
    else:
    	pass
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



import matplotlib.pyplot as plt 
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import (QMainWindow, QApplication, QWidget, QInputDialog, QLineEdit, QFileDialog, 
QDialog, QPushButton, QVBoxLayout, QComboBox, QCheckBox, QLabel,QAction, qApp, QTextEdit, QSpacerItem, QSizePolicy,QHBoxLayout, QGroupBox, QTableWidgetItem)
from PyQt5.QtCore import Qt, pyqtSignal, QObject, pyqtSlot
from PyQt5.QtGui import QIcon, QCursor


class PlotWidget(QWidget):
    def __init__(self, parent = None):
        QWidget.__init__(self, parent)
        self.canvas = FigureCanvas(Figure())
        vertical_layout = QVBoxLayout()
        vertical_layout.addWidget(self.canvas)
        self.canvas.axes = self.canvas.figure.add_subplot(111)
        self.setLayout(vertical_layout)


'''
FIXME: Obviously we need better tests than that..
'''
import sys
import unittest
import numpy as np
import scipy
# import matplotlib.pyplot as plt

from pysprint.core import evaluate
from pysprint.core.dataedits import find_peak
from pysprint import Generator, FFTMethod


class TestEvaluate(unittest.TestCase):

	def setUp(self):
		pass

	def tearDown(self):
		pass

	def test_min_max_bases(self):
		a = np.arange(100)
		b = np.arange(100)
		mins = [10,30,50,70,90]
		maxs = [20,40,60,80,100]
		disp, disp_s, fit = evaluate.min_max_method(a, b, [], [], 0, maxx=maxs, minx=mins, fitOrder=1, showGraph=False)
		np.testing.assert_almost_equal(disp[0], -0.3127272727301358)
		with self.assertRaises(ValueError):
			disp, disp_s, fit = evaluate.min_max_method([], b, [], [], 0, maxx=maxs, minx=mins, fitOrder=1, showGraph=False)
		with self.assertRaises(ValueError):
			disp, disp_s, fit = evaluate.min_max_method([], [], [], [], 0, maxx=maxs, minx=mins, fitOrder=1, showGraph=False)
		# with self.assertRaises(ValueError):
			# disp, disp_s, fit = evaluate.min_max_method(a, [], [], [], ref_point = 0, maxx=maxs, minx=mins, fitOrder=1, showGraph=False)
		with self.assertRaises(ValueError):
			disp, disp_s, fit = evaluate.min_max_method(a, b, [], [], 0, maxx=maxs, minx=mins, fitOrder=6, showGraph=False)
		# with self.assertRaises(np.core._exceptions.UFuncTypeError):
			# disp, disp_s, fit = evaluate.min_max_method(a, b, [], [], ref_point = 0, maxx='a', minx=mins, fitOrder=1, showGraph=False)
		# with self.assertRaises(ValueError):
			# disp, disp_s, fit = evaluate.min_max_method(a, b, b, b, ref_point=-50000, maxx=maxs, minx=mins, fitOrder=1, showGraph=False)

	def test_min_max_advanced(self):
		with self.assertRaises(ValueError):
			evaluate.min_max_method(np.arange(100), np.arange(100), [], [], ref_point = 10, fitOrder = 1, showGraph=False)
		a,b,c,d = np.loadtxt('test_arms.txt', delimiter = ',', unpack = True)
		maxs, _, mins, _ = find_peak(a,b,c,d, proMax=1, proMin=1, threshold=0.4)
		d1, d_s1, fit1 = evaluate.min_max_method(a,b,c,d, ref_point = 2.5, fitOrder = 1, maxx = maxs, minx = mins, showGraph = False)
		np.testing.assert_array_equal(d1[1:], [0,0,0,0])
		np.testing.assert_array_equal(d_s1[1:], [0,0,0,0])
		assert len(d1) == len(d_s1) == 5
		d2, d_s2, fit2 = evaluate.min_max_method(a,b,c,d, ref_point = 2.5, fitOrder = 2, maxx = maxs, minx = mins, showGraph = False)
		np.testing.assert_array_equal(d2[2:], [0,0,0])
		np.testing.assert_array_equal(d_s2[2:], [0,0,0])
		assert len(d2) == len(d_s2) == 5
		d3, d_s3, fit3 = evaluate.min_max_method(a,b,c,d, ref_point = 2.5, fitOrder = 3, maxx = maxs, minx = mins, showGraph = False)
		np.testing.assert_array_equal(d3[3:], [0,0])
		np.testing.assert_array_equal(d_s3[3:], [0,0])
		assert len(d3) == len(d_s3) == 5
		d4, d_s4, fit4 = evaluate.min_max_method(a,b,c,d, ref_point = 2.5, fitOrder = 4, maxx = maxs, minx = mins, showGraph = False)
		np.testing.assert_array_equal(d4[4:], [0])
		np.testing.assert_array_equal(d_s4[4:], [0])
		assert len(d4) == len(d_s4) == 5
		d5, d_s5, fit5 = evaluate.min_max_method(a,b,c,d, ref_point = 2.5, fitOrder = 5, maxx = maxs, minx = mins, showGraph = False)
		# np.testing.assert_array_equal(d3[3:], [0,0])
		# np.testing.assert_array_equal(d_s3[3:], [0,0])
		# plt.close('all')
		assert len(d5) == len(d_s5) == 5


			
	def test_cff(self):
		a = np.arange(100)
		b = np.arange(100)
		with self.assertRaises(TypeError):
			evaluate.cff_method(a, b, [], [], ref_point=0 , p0=[1, 1, 1, 1, 1, 1, 1, 1, 1])

	def test_ffts_primitive(self):
		#adapted from scipy's unittests
	    scipy.random.seed(1534)
	    x = scipy.randn(10) + 1j * scipy.randn(10)
	    fr, yf = evaluate.ifft_method(x, x, interpolate = False)
	    _, y = evaluate.fft_method(yf, yf)
	    np.testing.assert_allclose(y, x)


	def test_ffts_advanced(self):
		g = Generator(2.2, 2.8, 2.5, delay = 150000, GDD = -500, TOD = 3000, FOD = -20000, pulseWidth = 25)
		g.generate_freq()
		a,b = g.unpack()
		f = FFTMethod(a,b)
		f.ifft(interpolate = False)
		f.cut(126,30)
		f.fft()
		dis, _, _ = f.calculate(fit_order = 4, reference_point = 2.5)
		np.testing.assert_almost_equal(np.abs(dis[1]), 500.123, decimal = 2)
		np.testing.assert_almost_equal(np.abs(dis[2]), 3002.06, decimal = 2)
		np.testing.assert_almost_equal(np.abs(dis[3]), 19992.92, decimal = 2)
		np.testing.assert_almost_equal(dis[4], 0, decimal = 2)


	def test_windowing(self):
		a,b = np.loadtxt('test_window.txt', unpack = True, delimiter = ',')
		y_data = evaluate.cut_gaussian(a,b, 2.5, 0.2, 6)
		assert len(b) == len(y_data)
		np.testing.assert_almost_equal(y_data[0], 0)
		np.testing.assert_almost_equal(y_data[-1], 0)
		np.testing.assert_almost_equal(np.median(y_data), np.median(b), decimal = 2)

	def test_spp(self):
		pass

if __name__ == '__main__':
	unittest.main()


__all__ = ['ImportModel']

import numpy as np
from pysprint.core.loading import read_data

class InvalidCommand(Exception):
	pass


class ImportModel(object):
	def __init__(self, x, y, ref, sam):
		self.d = {}
		self.col1 = x
		self.col2 = y
		self.col3 = ref
		self.col4 = sam
		if len(self.col3) == 0 or len(self.col4) == 0:
			self._ncols = 2
		else:
			self._ncols = 4
		self.create_dict()

	def create_dict(self):
		# custom rules .. we might add more later.
		if len(self.col3) > 0 and len(self.col4) > 0:
			self.d['$y'] = '($2-$3-$4)/(2*np.sqrt($3*$4))'
		for i in range(self._ncols):
			self.d['chdomain($'+str(i+1)+')'] = '$' + str(i+1) + ' = np.pi*2*299.793/$' + str(i+1) 
			self.d['$'+str(i+1)] = 'self.col'  + str(i+1)
		self.d['^'] = '**'


	def exec_command(self, com):
		for command, executable in self.d.items():
		    com = com.replace(command, executable)
		try:
			exec(com)
		except SyntaxError:
			raise InvalidCommand

	def unpack(self):
		return self.col1, self.col2, self.col3, self.col4


from functools import wraps

import numpy as np


__all__ = ['scipy_disp', 'lmfit_disp', 'findNearest', 'find_closest',
           '_handle_input', 'print_disp']


def scipy_disp(r):
	for idx in range(len(r)):
		dispersion[idx] = dispersion[idx] * factorial(idx+1)
		dispersion_std[idx] = dispersion_std[idx] * factorial(idx+1)
	return dispersion, dispersion_std


def lmfit_disp(r):
	dispersion, dispersion_std = [], []
	for name, par in r:
		dispersion.append(par.value)
		dispersion_std.append(par.stderr)
	return dispersion, dispersion_std


def findNearest(array, value):
	#Finds the nearest element to the given value in the array
	#returns tuple: (element, element's index)
	
    array = np.asarray(array)
    idx = (np.abs(value - array)).argmin()
    return array[idx], idx


def _handle_input(initSpectrumX, initSpectrumY, referenceArmY, sampleArmY):
	"""
	Instead of handling the inputs in every function, there is this private method.

	Parameters
	----------

	initSpectrumX: array-like
	x-axis data

	initSpectrumY: array-like
	y-axis data

	referenceArmY, sampleArmY: array-like
	reference and sample arm spectrum evaluated at initSpectrumX

	Returns
	-------
	initSpectrumX: array-like
	unchanged x data

	Ydata: array-like
	the transformed y data

	"""
	if (len(initSpectrumX) > 0) and (len(referenceArmY) > 0) and (len(sampleArmY) > 0):
		Ydata = (initSpectrumY-referenceArmY-sampleArmY)/(2*np.sqrt(referenceArmY*sampleArmY))
	elif (len(referenceArmY) == 0) or (len(sampleArmY) == 0):
		Ydata = initSpectrumY
	elif len(initSpectrumX) == 0:
		raise ValueError('Please load the spectrum!\n')
	elif len(initSpectrumY) == 0:
		raise ValueError('Please load the spectrum!\n')
	else:
		raise TypeError('Input types are wrong.\n')
	return initSpectrumX,  Ydata


def find_closest(xValue, xArray, yArray):
	idx = (np.abs(xArray-xValue)).argmin()
	value = xArray[idx]
	return value, yArray[idx]


def print_disp(f):
    @wraps(f)
    def wrapping(*args, **kwargs):
        disp, disp_std, stri = f(*args, **kwargs)
        labels = ['GD', 'GDD','TOD', 'FOD', 'QOD']
        for i in range(len(labels)):
        	print(labels[i] + ' = ' + str(disp[i]) +  ' +/- ' + str(disp_std[i]) + ' fs^{}'.format(i+1))
        return disp, disp_std, stri
    return wrapping


"""
CURRENTLY POORLY WRITTEN, BUT IT WORKS.. I WILL UPDATE IT SOON.
"""
import sys
import numpy as np
from scipy.optimize import curve_fit
from pysprint.utils import findNearest as find_nearest
# from pysprint.logic import MainProgram

sys.path.append('..')


class FitOptimizer(object):
	"""Class to help achieve better fitting results."""
	def __init__(self, x, y, ref, sam, func=None, p0=None, outfunc=None):
		self.x = x
		self.y = y
		self.ref = ref
		self.sam = sam
		if not isinstance(self.x, np.ndarray):
			try:
				self.x = np.asarray(self.x)
			except:
				raise
		if not isinstance(self.y, np.ndarray):
			try:
				self.y = np.asarray(self.y)
			except:
				raise
		if not isinstance(self.ref, np.ndarray):
			try:
				self.ref = np.asarray(self.ref)
			except:
				pass
		if not isinstance(self.sam, np.ndarray):
			try:
				self.sam = np.asarray(self.sam)
			except:
				pass
		if len(self.ref) == 0:
			self._y_norm = self.y
		else:
			self._y_norm = (self.y - self.ref - self.sam)/(2*np.sqrt(self.sam*self.ref))
		self.func = func
		if p0 is None:
			self.p0 = []
		else:
			self.p0 = p0
		self.popt = self.p0
		self._init_set = False
		self.counter = 0

	def set_initial_region(self, percent, center):
		""" Determines the initial region to fit"""
		self._init_set = True
		_, idx = find_nearest(self.x, center)
		self._upper_bound = np.floor(idx + (percent/2)*(len(self.x) + 1))
		self._lower_bound = np.floor(idx - (percent/2)*(len(self.x) + 1))
		self._upper_bound = self._upper_bound.astype(int)
		self._lower_bound = self._lower_bound.astype(int)
		if self._lower_bound < 0:
			self._lower_bound = 0
		if self._upper_bound > len(self.x):
			self._upper_bound = len(self.x)
		self._x_curr = self.x[self._lower_bound:self._upper_bound]
		self._y_curr = self._y_norm[self._lower_bound:self._upper_bound]

	def _extend_region(self, extend_by=0.2):
		""" Extends region of fit"""
		self._new_lower = np.floor(self._lower_bound - extend_by*len(self.x))
		self._new_upper = np.floor(self._upper_bound + extend_by*len(self.x))
		self._new_lower = self._new_lower.astype(int)
		self._new_upper = self._new_upper.astype(int)
		self._lower_bound = self._new_lower
		self._upper_bound = self._new_upper
		if self._new_lower < 0:
			self._new_lower = 0
		if self._new_upper > len(self.x):
			self._new_upper = len(self.x)
		self._x_curr = self.x[self._new_lower:self._new_upper]
		self._y_curr = self._y_norm[self._new_lower:self._new_upper]

	def _make_fit(self):
		""" Makes fit  """
		try:
			if len(self._x_curr) == len(self.x):
				return True
			self.popt, self.pcov = curve_fit(
				self.func, self._x_curr, self._y_curr, maxfev=200000, p0=self.p0
				)
			self.p0 = self.popt
		except RuntimeError:
			if len(self.popt) > 4:
				self.p0[:3] = self.popt[:3] + np.random.normal(0, 100, len(self.popt)-3)
			else:
				self.p0 = self.popt + np.random.normal(0, 100, len(self.popt))
			self.popt, self.pcov = curve_fit(
				self.func, self._x_curr, self._y_curr, maxfev=200000, p0=self.p0
				)

	def _fit_goodness(self):
		residuals = self._y_curr - self.func(self._x_curr, *self.popt)
		ss_res = np.sum(residuals**2)
		ss_tot = np.sum((self._y_curr - np.mean(self._y_curr))**2)
		return 1 - (ss_res / ss_tot)

	def show_fit(self, obj=None):
		try:
			self.obj.axes.plot(
				self._x_curr, self._y_curr, 'k-', label='Affected data'
				)
			self.obj.axes.plot(
				self._x_curr, self.func(self._x_curr, *self.popt), 'r--', label='Fit'
				)
			self.obj.draw()
		except:
			pass

	def run_loop(
		self, r_extend_by, r_threshold, outfunc, max_tries=10000, show_steps=False
		):
		if self._init_set is False:
			raise ValueError('Set the initial conditions.')
		self._make_fit()
		while self._fit_goodness() > r_threshold:
			self._extend_region(r_extend_by)
			self._make_fit()
			self.counter += 1
			if self._make_fit() is True:
				self.show_fit(self.obj)
				outfunc('The params were:{}\n'.format(self.popt))
				return self.popt
			if self.counter == max_tries:
				self.show_fit(self.obj)
				outfunc('''Max tries ({}) reached.. try another initial params
					    .\n You can set the bounds at Edit --> Settings.'''.format(max_tries))
				return np.zeros_like(self.popt)

		while self._fit_goodness() < r_threshold:
			self._make_fit()
			self.counter += 1
			if self.counter == max_tries:
				self.show_fit(self.obj)
				outfunc('''Max tries ({}) reached.. try another initial params
					.\nYou can set the bounds at Edit --> Settings.'''.format(max_tries))
				return np.zeros_like(self.popt)
